/**
 * Copyright (C) 2002-2007 - INRIA (www.inria.fr)
 *
 * CAROL: Common Architecture for RMI ObjectWeb Layer
 *
 * This library is developed inside the OW2 Consortium,
 * http://www.ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JInterceptorStore.java 1318 2007-10-12 16:33:24Z loris $
 * --------------------------------------------------------------------------
 */

package org.ow2.carol.rmi.jrmp.interceptor.impl;

import java.rmi.server.UID;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Properties;

import org.ow2.carol.rmi.jrmp.interceptor.api.JInitInfo;
import org.ow2.carol.rmi.jrmp.interceptor.spi.JClientRequestInterceptor;
import org.ow2.carol.rmi.jrmp.interceptor.spi.JInitializer;
import org.ow2.carol.rmi.jrmp.interceptor.spi.JServerRequestInterceptor;
import org.ow2.carol.rmi.jrmp.interceptor.wrapper.JInitializerWrapper;
import org.ow2.carol.util.configuration.TraceCarol;

/**
 * Class <code>JInterceptorStore</code> is the CAROL JRMP Client and Server
 * Interceptors Storage System
 * @author Guillaume Riviere (Guillaume.Riviere@inrialpes.fr)
 * @version 1.0, 10/03/2003
 */
@SuppressWarnings("deprecation")
public class JInterceptorStore {

    /**
     * Initializer class prefix
     */
    public static final String INTIALIZER_PREFIX = "org.ow2.PortableInterceptor.JRMPInitializerClass";

    /**
     * private Interceptor for Context propagation
     */
    private static JServerRequestInterceptor[] sis = null;

    /**
     * private Interceptor for Context propagation
     */
    private static JClientRequestInterceptor[] cis = null;

    /**
     * private remote Interceptor cache for Context propagation
     */
    private static JClientRequestInterceptor[] rcis = null;

    /**
     * private remote UID for cache
     */
    private static UID uid = null;

    /**
     * private remote addr for cache
     */
    private static byte[] address = null;

    /**
     * private Interceptors Initializers for Context propagation
     */
    private static String[] initializers = null;

    /**
     * JRMPINfo Impl
     */
    private static JRMPInitInfoImpl jrmpInfo = new JRMPInitInfoImpl();

    /**
     * Initialize interceptors for a carol server
     */
    static {
        // Load the Interceptors
        try {
            JInitInfo jrmpInfo = new JRMPInitInfoImpl();
            String[] ins = getJRMPInitializers();
            for (String className : ins) {
                JInitializer jinit;
                Object obj = Thread.currentThread().getContextClassLoader().loadClass(className).newInstance();
                if(obj instanceof org.objectweb.carol.rmi.jrmp.interceptor.JInitializer) {
                    jinit = new JInitializerWrapper((org.objectweb.carol.rmi.jrmp.interceptor.JInitializer) obj);
                } else {
                    jinit = (JInitializer) obj;
                }
                jinit.pre_init(jrmpInfo);
                jinit.post_init(jrmpInfo);
            }
            sis = jrmpInfo.getServerRequestInterceptors();
            cis = jrmpInfo.getClientRequestInterceptors();
            // first remote reference = local reference
            rcis = cis;
            uid = JInterceptorHelper.getSpaceID();
            address = JInterceptorHelper.getInetAddress();
        } catch (Exception e) {
            //we did not found the interceptor do nothing but a trace ?
            TraceCarol.error("JrmpPRODelegate(), No interceptors found", e);
        }
    }

    /**
     * get the local server interceptor
     */
    public static JServerRequestInterceptor[] getLocalServerInterceptors() {
        return sis;
    }

    /**
     * get the local client interceptor
     */
    public static JClientRequestInterceptor[] getLocalClientInterceptors() {
        return cis;
    }

    /**
     * Get Initializers method
     * @return JRMP Initializers enumeration
     */
    public synchronized static String[] getJRMPInitializers() {
        // if exists, return it
        if (initializers != null) {
            return initializers;
        }
        // else, compute it.
        ArrayList<String> ins = new ArrayList<String>();
        Properties sys = System.getProperties();
        for (Enumeration<?> e = sys.propertyNames(); e.hasMoreElements();) {
            String pkey = (String) e.nextElement();
            if (pkey.startsWith(INTIALIZER_PREFIX)) {
                ins.add(pkey.substring(INTIALIZER_PREFIX.length() + 1));
            }
        }
        int sz = ins.size();
        initializers = new String[sz];
        return ins.toArray(initializers);
    }

    /**
     * Get interceptor if exists
     * @param raddr The remote address (later)
     * @param ruid The remote uid (later)
     * @param ia interceptors initializers
     * @return JClientRequestInterceptors [] , the interceptors
     */
    public synchronized static JClientRequestInterceptor[] setRemoteInterceptors(byte[] raddr, UID ruid, String[] ia) {
        if ((Arrays.equals(raddr, address)) && (ruid.equals(uid))) {
            return rcis;
        } else {
            jrmpInfo.clear();
            for (String className : ia) {
                JInitializer jinit;
                try {
                    Object obj = Class.forName(className).newInstance();
                    if(obj instanceof org.objectweb.carol.rmi.jrmp.interceptor.JInitializer) {
                        jinit = new JInitializerWrapper((org.objectweb.carol.rmi.jrmp.interceptor.JInitializer) obj);
                    } else {
                        jinit = (JInitializer) obj;
                    }
                    jinit.pre_init(jrmpInfo);
                    jinit.post_init(jrmpInfo);
                } catch (Exception e) {
                    TraceCarol.error("can not load interceptors", e);
                }
            }
            ruid = uid;
            address = raddr;
            rcis = jrmpInfo.getClientRequestInterceptors();
            return rcis;
        }
    }

}
